import os
import re
from typing import (
    Dict,
    List,
    Set,
    Tuple,
    Type,
)

from ..casing import safe_snake_case
from ..lib.google import protobuf as google_protobuf
from .naming import pythonize_class_name


WRAPPER_TYPES: Dict[str, Type] = {
    ".google.protobuf.DoubleValue": google_protobuf.DoubleValue,
    ".google.protobuf.FloatValue": google_protobuf.FloatValue,
    ".google.protobuf.Int32Value": google_protobuf.Int32Value,
    ".google.protobuf.Int64Value": google_protobuf.Int64Value,
    ".google.protobuf.UInt32Value": google_protobuf.UInt32Value,
    ".google.protobuf.UInt64Value": google_protobuf.UInt64Value,
    ".google.protobuf.BoolValue": google_protobuf.BoolValue,
    ".google.protobuf.StringValue": google_protobuf.StringValue,
    ".google.protobuf.BytesValue": google_protobuf.BytesValue,
}


def parse_source_type_name(field_type_name: str) -> Tuple[str, str]:
    """
    Split full source type name into package and type name.
    E.g. 'root.package.Message' -> ('root.package', 'Message')
         'root.Message.SomeEnum' -> ('root', 'Message.SomeEnum')
    """
    package_match = re.match(r"^\.?([^A-Z]+)\.(.+)", field_type_name)
    if package_match:
        package = package_match.group(1)
        name = package_match.group(2)
    else:
        package = ""
        name = field_type_name.lstrip(".")
    return package, name


def get_type_reference(
    *,
    package: str,
    imports: set,
    source_type: str,
    unwrap: bool = True,
    pydantic: bool = False,
) -> str:
    """
    Return a Python type name for a proto type reference. Adds the import if
    necessary. Unwraps well known type if required.
    """
    if unwrap:
        if source_type in WRAPPER_TYPES:
            wrapped_type = type(WRAPPER_TYPES[source_type]().value)
            return f"Optional[{wrapped_type.__name__}]"

        if source_type == ".google.protobuf.Duration":
            return "timedelta"

        elif source_type == ".google.protobuf.Timestamp":
            return "datetime"

    source_package, source_type = parse_source_type_name(source_type)

    current_package: List[str] = package.split(".") if package else []
    py_package: List[str] = source_package.split(".") if source_package else []
    py_type: str = pythonize_class_name(source_type)

    compiling_google_protobuf = current_package == ["google", "protobuf"]
    importing_google_protobuf = py_package == ["google", "protobuf"]
    if importing_google_protobuf and not compiling_google_protobuf:
        py_package = (
            ["aristaproto", "lib"] + (["pydantic"] if pydantic else []) + py_package
        )

    if py_package[:1] == ["aristaproto"]:
        return reference_absolute(imports, py_package, py_type)

    if py_package == current_package:
        return reference_sibling(py_type)

    if py_package[: len(current_package)] == current_package:
        return reference_descendent(current_package, imports, py_package, py_type)

    if current_package[: len(py_package)] == py_package:
        return reference_ancestor(current_package, imports, py_package, py_type)

    return reference_cousin(current_package, imports, py_package, py_type)


def reference_absolute(imports: Set[str], py_package: List[str], py_type: str) -> str:
    """
    Returns a reference to a python type located in the root, i.e. sys.path.
    """
    string_import = ".".join(py_package)
    string_alias = safe_snake_case(string_import)
    imports.add(f"import {string_import} as {string_alias}")
    return f'"{string_alias}.{py_type}"'


def reference_sibling(py_type: str) -> str:
    """
    Returns a reference to a python type within the same package as the current package.
    """
    return f'"{py_type}"'


def reference_descendent(
    current_package: List[str], imports: Set[str], py_package: List[str], py_type: str
) -> str:
    """
    Returns a reference to a python type in a package that is a descendent of the
    current package, and adds the required import that is aliased to avoid name
    conflicts.
    """
    importing_descendent = py_package[len(current_package) :]
    string_from = ".".join(importing_descendent[:-1])
    string_import = importing_descendent[-1]
    if string_from:
        string_alias = "_".join(importing_descendent)
        imports.add(f"from .{string_from} import {string_import} as {string_alias}")
        return f'"{string_alias}.{py_type}"'
    else:
        imports.add(f"from . import {string_import}")
        return f'"{string_import}.{py_type}"'


def reference_ancestor(
    current_package: List[str], imports: Set[str], py_package: List[str], py_type: str
) -> str:
    """
    Returns a reference to a python type in a package which is an ancestor to the
    current package, and adds the required import that is aliased (if possible) to avoid
    name conflicts.

    Adds trailing __ to avoid name mangling (python.org/dev/peps/pep-0008/#id34).
    """
    distance_up = len(current_package) - len(py_package)
    if py_package:
        string_import = py_package[-1]
        string_alias = f"_{'_' * distance_up}{string_import}__"
        string_from = f"..{'.' * distance_up}"
        imports.add(f"from {string_from} import {string_import} as {string_alias}")
        return f'"{string_alias}.{py_type}"'
    else:
        string_alias = f"{'_' * distance_up}{py_type}__"
        imports.add(f"from .{'.' * distance_up} import {py_type} as {string_alias}")
        return f'"{string_alias}"'


def reference_cousin(
    current_package: List[str], imports: Set[str], py_package: List[str], py_type: str
) -> str:
    """
    Returns a reference to a python type in a package that is not descendent, ancestor
    or sibling, and adds the required import that is aliased to avoid name conflicts.
    """
    shared_ancestry = os.path.commonprefix([current_package, py_package])  # type: ignore
    distance_up = len(current_package) - len(shared_ancestry)
    string_from = f".{'.' * distance_up}" + ".".join(
        py_package[len(shared_ancestry) : -1]
    )
    string_import = py_package[-1]
    # Add trailing __ to avoid name mangling (python.org/dev/peps/pep-0008/#id34)
    string_alias = (
        f"{'_' * distance_up}"
        + safe_snake_case(".".join(py_package[len(shared_ancestry) :]))
        + "__"
    )
    imports.add(f"from {string_from} import {string_import} as {string_alias}")
    return f'"{string_alias}.{py_type}"'



from tests.output_aristaproto.regression_387 import (
    ParentElement,
    Test,
)


def test_regression_387():
    el = ParentElement(name="test", elems=[Test(id=0), Test(id=42)])
    binary = bytes(el)
    decoded = ParentElement().parse(binary)
    assert decoded == el
    assert decoded.elems == [Test(id=0), Test(id=42)]



"""Plugin model dataclasses.

These classes are meant to be an intermediate representation
of protobuf objects. They are used to organize the data collected during parsing.

The general intention is to create a doubly-linked tree-like structure
with the following types of references:
- Downwards references: from message -> fields, from output package -> messages
or from service -> service methods
- Upwards references: from field -> message, message -> package.
- Input/output message references: from a service method to it's corresponding
input/output messages, which may even be in another package.

There are convenience methods to allow climbing up and down this tree, for
example to retrieve the list of all messages that are in the same package as
the current message.

Most of these classes take as inputs:
- proto_obj: A reference to it's corresponding protobuf object as
presented by the protoc plugin.
- parent: a reference to the parent object in the tree.

With this information, the class is able to expose attributes,
such as a pythonized name, that will be calculated from proto_obj.

The instantiation should also attach a reference to the new object
into the corresponding place within it's parent object. For example,
instantiating field `A` with parent message `B` should add a
reference to `A` to `B`'s `fields` attribute.
"""


import builtins
import re
import textwrap
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Type,
    Union,
)

import aristaproto
from aristaproto import which_one_of
from aristaproto.casing import sanitize_name
from aristaproto.compile.importing import (
    get_type_reference,
    parse_source_type_name,
)
from aristaproto.compile.naming import (
    pythonize_class_name,
    pythonize_field_name,
    pythonize_method_name,
)
from aristaproto.lib.google.protobuf import (
    DescriptorProto,
    EnumDescriptorProto,
    Field,
    FieldDescriptorProto,
    FieldDescriptorProtoLabel,
    FieldDescriptorProtoType,
    FileDescriptorProto,
    MethodDescriptorProto,
)
from aristaproto.lib.google.protobuf.compiler import CodeGeneratorRequest

from ..compile.importing import (
    get_type_reference,
    parse_source_type_name,
)
from ..compile.naming import (
    pythonize_class_name,
    pythonize_enum_member_name,
    pythonize_field_name,
    pythonize_method_name,
)


# Create a unique placeholder to deal with
# https://stackoverflow.com/questions/51575931/class-inheritance-in-python-3-7-dataclasses
PLACEHOLDER = object()

# Organize proto types into categories
PROTO_FLOAT_TYPES = (
    FieldDescriptorProtoType.TYPE_DOUBLE,  # 1
    FieldDescriptorProtoType.TYPE_FLOAT,  # 2
)
PROTO_INT_TYPES = (
    FieldDescriptorProtoType.TYPE_INT64,  # 3
    FieldDescriptorProtoType.TYPE_UINT64,  # 4
    FieldDescriptorProtoType.TYPE_INT32,  # 5
    FieldDescriptorProtoType.TYPE_FIXED64,  # 6
    FieldDescriptorProtoType.TYPE_FIXED32,  # 7
    FieldDescriptorProtoType.TYPE_UINT32,  # 13
    FieldDescriptorProtoType.TYPE_SFIXED32,  # 15
    FieldDescriptorProtoType.TYPE_SFIXED64,  # 16
    FieldDescriptorProtoType.TYPE_SINT32,  # 17
    FieldDescriptorProtoType.TYPE_SINT64,  # 18
)
PROTO_BOOL_TYPES = (FieldDescriptorProtoType.TYPE_BOOL,)  # 8
PROTO_STR_TYPES = (FieldDescriptorProtoType.TYPE_STRING,)  # 9
PROTO_BYTES_TYPES = (FieldDescriptorProtoType.TYPE_BYTES,)  # 12
PROTO_MESSAGE_TYPES = (
    FieldDescriptorProtoType.TYPE_MESSAGE,  # 11
    FieldDescriptorProtoType.TYPE_ENUM,  # 14
)
PROTO_MAP_TYPES = (FieldDescriptorProtoType.TYPE_MESSAGE,)  # 11
PROTO_PACKED_TYPES = (
    FieldDescriptorProtoType.TYPE_DOUBLE,  # 1
    FieldDescriptorProtoType.TYPE_FLOAT,  # 2
    FieldDescriptorProtoType.TYPE_INT64,  # 3
    FieldDescriptorProtoType.TYPE_UINT64,  # 4
    FieldDescriptorProtoType.TYPE_INT32,  # 5
    FieldDescriptorProtoType.TYPE_FIXED64,  # 6
    FieldDescriptorProtoType.TYPE_FIXED32,  # 7
    FieldDescriptorProtoType.TYPE_BOOL,  # 8
    FieldDescriptorProtoType.TYPE_UINT32,  # 13
    FieldDescriptorProtoType.TYPE_SFIXED32,  # 15
    FieldDescriptorProtoType.TYPE_SFIXED64,  # 16
    FieldDescriptorProtoType.TYPE_SINT32,  # 17
    FieldDescriptorProtoType.TYPE_SINT64,  # 18
)


def monkey_patch_oneof_index():
    """
    The compiler message types are written for proto2, but we read them as proto3.
    For this to work in the case of the oneof_index fields, which depend on being able
    to tell whether they were set, we have to treat them as oneof fields. This method
    monkey patches the generated classes after the fact to force this behaviour.
    """
    object.__setattr__(
        FieldDescriptorProto.__dataclass_fields__["oneof_index"].metadata[
            "aristaproto"
        ],
        "group",
        "oneof_index",
    )
    object.__setattr__(
        Field.__dataclass_fields__["oneof_index"].metadata["aristaproto"],
        "group",
        "oneof_index",
    )


def get_comment(
    proto_file: "FileDescriptorProto", path: List[int], indent: int = 4
) -> str:
    pad = " " * indent
    for sci_loc in proto_file.source_code_info.location:
        if list(sci_loc.path) == path and sci_loc.leading_comments:
            lines = sci_loc.leading_comments.strip().replace("\t", "    ").split("\n")
            # This is a field, message, enum, service, or method
            if len(lines) == 1 and len(lines[0]) < 79 - indent - 6:
                lines[0] = lines[0].strip('"')
                # rstrip to remove trailing spaces including whitespaces from empty lines.
                return f'{pad}"""{lines[0]}"""'
            else:
                # rstrip to remove trailing spaces including empty lines.
                padded = [f"\n{pad}{line}".rstrip(" ") for line in lines]
                joined = "".join(padded)
                return f'{pad}"""{joined}\n{pad}"""'

    return ""


class ProtoContentBase:
    """Methods common to MessageCompiler, ServiceCompiler and ServiceMethodCompiler."""

    source_file: FileDescriptorProto
    path: List[int]
    comment_indent: int = 4
    parent: Union["aristaproto.Message", "OutputTemplate"]

    __dataclass_fields__: Dict[str, object]

    def __post_init__(self) -> None:
        """Checks that no fake default fields were left as placeholders."""
        for field_name, field_val in self.__dataclass_fields__.items():
            if field_val is PLACEHOLDER:
                raise ValueError(f"`{field_name}` is a required field.")

    @property
    def output_file(self) -> "OutputTemplate":
        current = self
        while not isinstance(current, OutputTemplate):
            current = current.parent
        return current

    @property
    def request(self) -> "PluginRequestCompiler":
        current = self
        while not isinstance(current, OutputTemplate):
            current = current.parent
        return current.parent_request

    @property
    def comment(self) -> str:
        """Crawl the proto source code and retrieve comments
        for this object.
        """
        return get_comment(
            proto_file=self.source_file, path=self.path, indent=self.comment_indent
        )


@dataclass
class PluginRequestCompiler:
    plugin_request_obj: CodeGeneratorRequest
    output_packages: Dict[str, "OutputTemplate"] = field(default_factory=dict)

    @property
    def all_messages(self) -> List["MessageCompiler"]:
        """All of the messages in this request.

        Returns
        -------
        List[MessageCompiler]
            List of all of the messages in this request.
        """
        return [
            msg for output in self.output_packages.values() for msg in output.messages
        ]


@dataclass
class OutputTemplate:
    """Representation of an output .py file.

    Each output file corresponds to a .proto input file,
    but may need references to other .proto files to be
    built.
    """

    parent_request: PluginRequestCompiler
    package_proto_obj: FileDescriptorProto
    input_files: List[str] = field(default_factory=list)
    imports: Set[str] = field(default_factory=set)
    datetime_imports: Set[str] = field(default_factory=set)
    typing_imports: Set[str] = field(default_factory=set)
    pydantic_imports: Set[str] = field(default_factory=set)
    builtins_import: bool = False
    messages: List["MessageCompiler"] = field(default_factory=list)
    enums: List["EnumDefinitionCompiler"] = field(default_factory=list)
    services: List["ServiceCompiler"] = field(default_factory=list)
    imports_type_checking_only: Set[str] = field(default_factory=set)
    pydantic_dataclasses: bool = False
    output: bool = True

    @property
    def package(self) -> str:
        """Name of input package.

        Returns
        -------
        str
            Name of input package.
        """
        return self.package_proto_obj.package

    @property
    def input_filenames(self) -> Iterable[str]:
        """Names of the input files used to build this output.

        Returns
        -------
        Iterable[str]
            Names of the input files used to build this output.
        """
        return sorted(f.name for f in self.input_files)

    @property
    def python_module_imports(self) -> Set[str]:
        imports = set()
        if any(x for x in self.messages if any(x.deprecated_fields)):
            imports.add("warnings")
        if self.builtins_import:
            imports.add("builtins")
        return imports


@dataclass
class MessageCompiler(ProtoContentBase):
    """Representation of a protobuf message."""

    source_file: FileDescriptorProto
    parent: Union["MessageCompiler", OutputTemplate] = PLACEHOLDER
    proto_obj: DescriptorProto = PLACEHOLDER
    path: List[int] = PLACEHOLDER
    fields: List[Union["FieldCompiler", "MessageCompiler"]] = field(
        default_factory=list
    )
    deprecated: bool = field(default=False, init=False)
    builtins_types: Set[str] = field(default_factory=set)

    def __post_init__(self) -> None:
        # Add message to output file
        if isinstance(self.parent, OutputTemplate):
            if isinstance(self, EnumDefinitionCompiler):
                self.output_file.enums.append(self)
            else:
                self.output_file.messages.append(self)
        self.deprecated = self.proto_obj.options.deprecated
        super().__post_init__()

    @property
    def proto_name(self) -> str:
        return self.proto_obj.name

    @property
    def py_name(self) -> str:
        return pythonize_class_name(self.proto_name)

    @property
    def annotation(self) -> str:
        if self.repeated:
            return f"List[{self.py_name}]"
        return self.py_name

    @property
    def deprecated_fields(self) -> Iterator[str]:
        for f in self.fields:
            if f.deprecated:
                yield f.py_name

    @property
    def has_deprecated_fields(self) -> bool:
        return any(self.deprecated_fields)

    @property
    def has_oneof_fields(self) -> bool:
        return any(isinstance(field, OneOfFieldCompiler) for field in self.fields)

    @property
    def has_message_field(self) -> bool:
        return any(
            (
                field.proto_obj.type in PROTO_MESSAGE_TYPES
                for field in self.fields
                if isinstance(field.proto_obj, FieldDescriptorProto)
            )
        )


def is_map(
    proto_field_obj: FieldDescriptorProto, parent_message: DescriptorProto
) -> bool:
    """True if proto_field_obj is a map, otherwise False."""
    if proto_field_obj.type == FieldDescriptorProtoType.TYPE_MESSAGE:
        if not hasattr(parent_message, "nested_type"):
            return False

        # This might be a map...
        message_type = proto_field_obj.type_name.split(".").pop().lower()
        map_entry = f"{proto_field_obj.name.replace('_', '').lower()}entry"
        if message_type == map_entry:
            for nested in parent_message.nested_type:  # parent message
                if (
                    nested.name.replace("_", "").lower() == map_entry
                    and nested.options.map_entry
                ):
                    return True
    return False


def is_oneof(proto_field_obj: FieldDescriptorProto) -> bool:
    """
    True if proto_field_obj is a OneOf, otherwise False.

    .. warning::
        Becuase the message from protoc is defined in proto2, and aristaproto works with
        proto3, and interpreting the FieldDescriptorProto.oneof_index field requires
        distinguishing between default and unset values (which proto3 doesn't support),
        we have to hack the generated FieldDescriptorProto class for this to work.
        The hack consists of setting group="oneof_index" in the field metadata,
        essentially making oneof_index the sole member of a one_of group, which allows
        us to tell whether it was set, via the which_one_of interface.
    """

    return (
        not proto_field_obj.proto3_optional
        and which_one_of(proto_field_obj, "oneof_index")[0] == "oneof_index"
    )


@dataclass
class FieldCompiler(MessageCompiler):
    parent: MessageCompiler = PLACEHOLDER
    proto_obj: FieldDescriptorProto = PLACEHOLDER

    def __post_init__(self) -> None:
        # Add field to message
        self.parent.fields.append(self)
        # Check for new imports
        self.add_imports_to(self.output_file)
        super().__post_init__()  # call FieldCompiler-> MessageCompiler __post_init__

    def get_field_string(self, indent: int = 4) -> str:
        """Construct string representation of this field as a field."""
        name = f"{self.py_name}"
        annotations = f": {self.annotation}"
        field_args = ", ".join(
            ([""] + self.aristaproto_field_args) if self.aristaproto_field_args else []
        )
        aristaproto_field_type = (
            f"aristaproto.{self.field_type}_field({self.proto_obj.number}{field_args})"
        )
        if self.py_name in dir(builtins):
            self.parent.builtins_types.add(self.py_name)
        return f"{name}{annotations} = {aristaproto_field_type}"

    @property
    def aristaproto_field_args(self) -> List[str]:
        args = []
        if self.field_wraps:
            args.append(f"wraps={self.field_wraps}")
        if self.optional:
            args.append(f"optional=True")
        return args

    @property
    def datetime_imports(self) -> Set[str]:
        imports = set()
        annotation = self.annotation
        # FIXME: false positives - e.g. `MyDatetimedelta`
        if "timedelta" in annotation:
            imports.add("timedelta")
        if "datetime" in annotation:
            imports.add("datetime")
        return imports

    @property
    def typing_imports(self) -> Set[str]:
        imports = set()
        annotation = self.annotation
        if "Optional[" in annotation:
            imports.add("Optional")
        if "List[" in annotation:
            imports.add("List")
        if "Dict[" in annotation:
            imports.add("Dict")
        return imports

    @property
    def pydantic_imports(self) -> Set[str]:
        return set()

    @property
    def use_builtins(self) -> bool:
        return self.py_type in self.parent.builtins_types or (
            self.py_type == self.py_name and self.py_name in dir(builtins)
        )

    def add_imports_to(self, output_file: OutputTemplate) -> None:
        output_file.datetime_imports.update(self.datetime_imports)
        output_file.typing_imports.update(self.typing_imports)
        output_file.pydantic_imports.update(self.pydantic_imports)
        output_file.builtins_import = output_file.builtins_import or self.use_builtins

    @property
    def field_wraps(self) -> Optional[str]:
        """Returns aristaproto wrapped field type or None."""
        match_wrapper = re.match(
            r"\.google\.protobuf\.(.+)Value$", self.proto_obj.type_name
        )
        if match_wrapper:
            wrapped_type = "TYPE_" + match_wrapper.group(1).upper()
            if hasattr(aristaproto, wrapped_type):
                return f"aristaproto.{wrapped_type}"
        return None

    @property
    def repeated(self) -> bool:
        return (
            self.proto_obj.label == FieldDescriptorProtoLabel.LABEL_REPEATED
            and not is_map(self.proto_obj, self.parent)
        )

    @property
    def optional(self) -> bool:
        return self.proto_obj.proto3_optional

    @property
    def mutable(self) -> bool:
        """True if the field is a mutable type, otherwise False."""
        return self.annotation.startswith(("List[", "Dict["))

    @property
    def field_type(self) -> str:
        """String representation of proto field type."""
        return (
            FieldDescriptorProtoType(self.proto_obj.type)
            .name.lower()
            .replace("type_", "")
        )

    @property
    def default_value_string(self) -> str:
        """Python representation of the default proto value."""
        if self.repeated:
            return "[]"
        if self.optional:
            return "None"
        if self.py_type == "int":
            return "0"
        if self.py_type == "float":
            return "0.0"
        elif self.py_type == "bool":
            return "False"
        elif self.py_type == "str":
            return '""'
        elif self.py_type == "bytes":
            return 'b""'
        elif self.field_type == "enum":
            enum_proto_obj_name = self.proto_obj.type_name.split(".").pop()
            enum = next(
                e
                for e in self.output_file.enums
                if e.proto_obj.name == enum_proto_obj_name
            )
            return enum.default_value_string
        else:
            # Message type
            return "None"

    @property
    def packed(self) -> bool:
        """True if the wire representation is a packed format."""
        return self.repeated and self.proto_obj.type in PROTO_PACKED_TYPES

    @property
    def py_name(self) -> str:
        """Pythonized name."""
        return pythonize_field_name(self.proto_name)

    @property
    def proto_name(self) -> str:
        """Original protobuf name."""
        return self.proto_obj.name

    @property
    def py_type(self) -> str:
        """String representation of Python type."""
        if self.proto_obj.type in PROTO_FLOAT_TYPES:
            return "float"
        elif self.proto_obj.type in PROTO_INT_TYPES:
            return "int"
        elif self.proto_obj.type in PROTO_BOOL_TYPES:
            return "bool"
        elif self.proto_obj.type in PROTO_STR_TYPES:
            return "str"
        elif self.proto_obj.type in PROTO_BYTES_TYPES:
            return "bytes"
        elif self.proto_obj.type in PROTO_MESSAGE_TYPES:
            # Type referencing another defined Message or a named enum
            return get_type_reference(
                package=self.output_file.package,
                imports=self.output_file.imports,
                source_type=self.proto_obj.type_name,
                pydantic=self.output_file.pydantic_dataclasses,
            )
        else:
            raise NotImplementedError(f"Unknown type {self.proto_obj.type}")

    @property
    def annotation(self) -> str:
        py_type = self.py_type
        if self.use_builtins:
            py_type = f"builtins.{py_type}"
        if self.repeated:
            return f"List[{py_type}]"
        if self.optional:
            return f"Optional[{py_type}]"
        return py_type


@dataclass
class OneOfFieldCompiler(FieldCompiler):
    @property
    def aristaproto_field_args(self) -> List[str]:
        args = super().aristaproto_field_args
        group = self.parent.proto_obj.oneof_decl[self.proto_obj.oneof_index].name
        args.append(f'group="{group}"')
        return args


@dataclass
class PydanticOneOfFieldCompiler(OneOfFieldCompiler):
    @property
    def optional(self) -> bool:
        # Force the optional to be True. This will allow the pydantic dataclass
        # to validate the object correctly by allowing the field to be let empty.
        # We add a pydantic validator later to ensure exactly one field is defined.
        return True

    @property
    def pydantic_imports(self) -> Set[str]:
        return {"root_validator"}


@dataclass
class MapEntryCompiler(FieldCompiler):
    py_k_type: Type = PLACEHOLDER
    py_v_type: Type = PLACEHOLDER
    proto_k_type: str = PLACEHOLDER
    proto_v_type: str = PLACEHOLDER

    def __post_init__(self) -> None:
        """Explore nested types and set k_type and v_type if unset."""
        map_entry = f"{self.proto_obj.name.replace('_', '').lower()}entry"
        for nested in self.parent.proto_obj.nested_type:
            if (
                nested.name.replace("_", "").lower() == map_entry
                and nested.options.map_entry
            ):
                # Get Python types
                self.py_k_type = FieldCompiler(
                    source_file=self.source_file,
                    parent=self,
                    proto_obj=nested.field[0],  # key
                ).py_type
                self.py_v_type = FieldCompiler(
                    source_file=self.source_file,
                    parent=self,
                    proto_obj=nested.field[1],  # value
                ).py_type

                # Get proto types
                self.proto_k_type = FieldDescriptorProtoType(nested.field[0].type).name
                self.proto_v_type = FieldDescriptorProtoType(nested.field[1].type).name
        super().__post_init__()  # call FieldCompiler-> MessageCompiler __post_init__

    @property
    def aristaproto_field_args(self) -> List[str]:
        return [f"aristaproto.{self.proto_k_type}", f"aristaproto.{self.proto_v_type}"]

    @property
    def field_type(self) -> str:
        return "map"

    @property
    def annotation(self) -> str:
        return f"Dict[{self.py_k_type}, {self.py_v_type}]"

    @property
    def repeated(self) -> bool:
        return False  # maps cannot be repeated


@dataclass
class EnumDefinitionCompiler(MessageCompiler):
    """Representation of a proto Enum definition."""

    proto_obj: EnumDescriptorProto = PLACEHOLDER
    entries: List["EnumDefinitionCompiler.EnumEntry"] = PLACEHOLDER

    @dataclass(unsafe_hash=True)
    class EnumEntry:
        """Representation of an Enum entry."""

        name: str
        value: int
        comment: str

    def __post_init__(self) -> None:
        # Get entries/allowed values for this Enum
        self.entries = [
            self.EnumEntry(
                name=pythonize_enum_member_name(
                    entry_proto_value.name, self.proto_obj.name
                ),
                value=entry_proto_value.number,
                comment=get_comment(
                    proto_file=self.source_file, path=self.path + [2, entry_number]
                ),
            )
            for entry_number, entry_proto_value in enumerate(self.proto_obj.value)
        ]
        super().__post_init__()  # call MessageCompiler __post_init__

    @property
    def default_value_string(self) -> str:
        """Python representation of the default value for Enums.

        As per the spec, this is the first value of the Enum.
        """
        return str(self.entries[0].value)  # ideally, should ALWAYS be int(0)!


@dataclass
class ServiceCompiler(ProtoContentBase):
    parent: OutputTemplate = PLACEHOLDER
    proto_obj: DescriptorProto = PLACEHOLDER
    path: List[int] = PLACEHOLDER
    methods: List["ServiceMethodCompiler"] = field(default_factory=list)

    def __post_init__(self) -> None:
        # Add service to output file
        self.output_file.services.append(self)
        self.output_file.typing_imports.add("Dict")
        super().__post_init__()  # check for unset fields

    @property
    def proto_name(self) -> str:
        return self.proto_obj.name

    @property
    def py_name(self) -> str:
        return pythonize_class_name(self.proto_name)


@dataclass
class ServiceMethodCompiler(ProtoContentBase):
    parent: ServiceCompiler
    proto_obj: MethodDescriptorProto
    path: List[int] = PLACEHOLDER
    comment_indent: int = 8

    def __post_init__(self) -> None:
        # Add method to service
        self.parent.methods.append(self)

        # Check for imports
        if "Optional" in self.py_output_message_type:
            self.output_file.typing_imports.add("Optional")

        # Check for Async imports
        if self.client_streaming:
            self.output_file.typing_imports.add("AsyncIterable")
            self.output_file.typing_imports.add("Iterable")
            self.output_file.typing_imports.add("Union")

        # Required by both client and server
        if self.client_streaming or self.server_streaming:
            self.output_file.typing_imports.add("AsyncIterator")

        # add imports required for request arguments timeout, deadline and metadata
        self.output_file.typing_imports.add("Optional")
        self.output_file.imports_type_checking_only.add("import grpclib.server")
        self.output_file.imports_type_checking_only.add(
            "from aristaproto.grpc.grpclib_client import MetadataLike"
        )
        self.output_file.imports_type_checking_only.add(
            "from grpclib.metadata import Deadline"
        )

        super().__post_init__()  # check for unset fields

    @property
    def py_name(self) -> str:
        """Pythonized method name."""
        return pythonize_method_name(self.proto_obj.name)

    @property
    def proto_name(self) -> str:
        """Original protobuf name."""
        return self.proto_obj.name

    @property
    def route(self) -> str:
        package_part = (
            f"{self.output_file.package}." if self.output_file.package else ""
        )
        return f"/{package_part}{self.parent.proto_name}/{self.proto_name}"

    @property
    def py_input_message(self) -> Optional[MessageCompiler]:
        """Find the input message object.

        Returns
        -------
        Optional[MessageCompiler]
            Method instance representing the input message.
            If not input message could be found or there are no
            input messages, None is returned.
        """
        package, name = parse_source_type_name(self.proto_obj.input_type)

        # Nested types are currently flattened without dots.
        # Todo: keep a fully quantified name in types, that is
        # comparable with method.input_type
        for msg in self.request.all_messages:
            if (
                msg.py_name == pythonize_class_name(name.replace(".", ""))
                and msg.output_file.package == package
            ):
                return msg
        return None

    @property
    def py_input_message_type(self) -> str:
        """String representation of the Python type corresponding to the
        input message.

        Returns
        -------
        str
            String representation of the Python type corresponding to the input message.
        """
        return get_type_reference(
            package=self.output_file.package,
            imports=self.output_file.imports,
            source_type=self.proto_obj.input_type,
            unwrap=False,
            pydantic=self.output_file.pydantic_dataclasses,
        ).strip('"')

    @property
    def py_input_message_param(self) -> str:
        """Param name corresponding to py_input_message_type.

        Returns
        -------
        str
            Param name corresponding to py_input_message_type.
        """
        return pythonize_field_name(self.py_input_message_type)

    @property
    def py_output_message_type(self) -> str:
        """String representation of the Python type corresponding to the
        output message.

        Returns
        -------
        str
            String representation of the Python type corresponding to the output message.
        """
        return get_type_reference(
            package=self.output_file.package,
            imports=self.output_file.imports,
            source_type=self.proto_obj.output_type,
            unwrap=False,
            pydantic=self.output_file.pydantic_dataclasses,
        ).strip('"')

    @property
    def client_streaming(self) -> bool:
        return self.proto_obj.client_streaming

    @property
    def server_streaming(self) -> bool:
        return self.proto_obj.server_streaming



from tests.output_aristaproto.mapmessage import (
    Nested,
    Test,
)


def test_mapmessage_to_dict_preserves_message():
    message = Test(
        items={
            "test": Nested(
                count=1,
            )
        }
    )

    message.to_dict()

    assert isinstance(message.items["test"], Nested), "Wrong nested type after to_dict"



from aristaproto import casing


def pythonize_class_name(name: str) -> str:
    return casing.pascal_case(name)


def pythonize_field_name(name: str) -> str:
    return casing.safe_snake_case(name)


def pythonize_method_name(name: str) -> str:
    return casing.safe_snake_case(name)


def pythonize_enum_member_name(name: str, enum_name: str) -> str:
    enum_name = casing.snake_case(enum_name).upper()
    find = name.find(enum_name)
    if find != -1:
        name = name[find + len(enum_name) :].strip("_")
    return casing.sanitize_name(name)



from typing import Dict

import grpclib
import grpclib.server

from tests.output_aristaproto.service import (
    DoThingRequest,
    DoThingResponse,
    GetThingRequest,
    GetThingResponse,
)


class ThingService:
    def __init__(self, test_hook=None):
        # This lets us pass assertions to the servicer ;)
        self.test_hook = test_hook

    async def do_thing(
        self, stream: "grpclib.server.Stream[DoThingRequest, DoThingResponse]"
    ):
        request = await stream.recv_message()
        if self.test_hook is not None:
            self.test_hook(stream)
        await stream.send_message(DoThingResponse([request.name]))

    async def do_many_things(
        self, stream: "grpclib.server.Stream[DoThingRequest, DoThingResponse]"
    ):
        thing_names = [request.name async for request in stream]
        if self.test_hook is not None:
            self.test_hook(stream)
        await stream.send_message(DoThingResponse(thing_names))

    async def get_thing_versions(
        self, stream: "grpclib.server.Stream[GetThingRequest, GetThingResponse]"
    ):
        request = await stream.recv_message()
        if self.test_hook is not None:
            self.test_hook(stream)
        for version_num in range(1, 6):
            await stream.send_message(
                GetThingResponse(name=request.name, version=version_num)
            )

    async def get_different_things(
        self, stream: "grpclib.server.Stream[GetThingRequest, GetThingResponse]"
    ):
        if self.test_hook is not None:
            self.test_hook(stream)
        #  Respond to each input item immediately
        response_num = 0
        async for request in stream:
            response_num += 1
            await stream.send_message(
                GetThingResponse(name=request.name, version=response_num)
            )

    def __mapping__(self) -> Dict[str, "grpclib.const.Handler"]:
        return {
            "/service.Test/DoThing": grpclib.const.Handler(
                self.do_thing,
                grpclib.const.Cardinality.UNARY_UNARY,
                DoThingRequest,
                DoThingResponse,
            ),
            "/service.Test/DoManyThings": grpclib.const.Handler(
                self.do_many_things,
                grpclib.const.Cardinality.STREAM_UNARY,
                DoThingRequest,
                DoThingResponse,
            ),
            "/service.Test/GetThingVersions": grpclib.const.Handler(
                self.get_thing_versions,
                grpclib.const.Cardinality.UNARY_STREAM,
                GetThingRequest,
                GetThingResponse,
            ),
            "/service.Test/GetDifferentThings": grpclib.const.Handler(
                self.get_different_things,
                grpclib.const.Cardinality.STREAM_STREAM,
                GetThingRequest,
                GetThingResponse,
            ),
        }



import copy
import sys

import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--repeat", type=int, default=1, help="repeat the operation multiple times"
    )


@pytest.fixture(scope="session")
def repeat(request):
    return request.config.getoption("repeat")


@pytest.fixture
def reset_sys_path():
    original = copy.deepcopy(sys.path)
    yield
    sys.path = original



from typing import (
    AsyncIterable,
    AsyncIterator,
)

import pytest
from grpclib.testing import ChannelFor

from tests.output_aristaproto.example_service import (
    ExampleRequest,
    ExampleResponse,
    TestBase,
    TestStub,
)


class ExampleService(TestBase):
    async def example_unary_unary(
        self, example_request: ExampleRequest
    ) -> "ExampleResponse":
        return ExampleResponse(
            example_string=example_request.example_string,
            example_integer=example_request.example_integer,
        )

    async def example_unary_stream(
        self, example_request: ExampleRequest
    ) -> AsyncIterator["ExampleResponse"]:
        response = ExampleResponse(
            example_string=example_request.example_string,
            example_integer=example_request.example_integer,
        )
        yield response
        yield response
        yield response

    async def example_stream_unary(
        self, example_request_iterator: AsyncIterator["ExampleRequest"]
    ) -> "ExampleResponse":
        async for example_request in example_request_iterator:
            return ExampleResponse(
                example_string=example_request.example_string,
                example_integer=example_request.example_integer,
            )

    async def example_stream_stream(
        self, example_request_iterator: AsyncIterator["ExampleRequest"]
    ) -> AsyncIterator["ExampleResponse"]:
        async for example_request in example_request_iterator:
            yield ExampleResponse(
                example_string=example_request.example_string,
                example_integer=example_request.example_integer,
            )


@pytest.mark.asyncio
async def test_calls_with_different_cardinalities():
    example_request = ExampleRequest("test string", 42)

    async with ChannelFor([ExampleService()]) as channel:
        stub = TestStub(channel)

        # unary unary
        response = await stub.example_unary_unary(example_request)
        assert response.example_string == example_request.example_string
        assert response.example_integer == example_request.example_integer

        # unary stream
        async for response in stub.example_unary_stream(example_request):
            assert response.example_string == example_request.example_string
            assert response.example_integer == example_request.example_integer

        # stream unary
        async def request_iterator():
            yield example_request
            yield example_request
            yield example_request

        response = await stub.example_stream_unary(request_iterator())
        assert response.example_string == example_request.example_string
        assert response.example_integer == example_request.example_integer

        # stream stream
        async for response in stub.example_stream_stream(request_iterator()):
            assert response.example_string == example_request.example_string
            assert response.example_integer == example_request.example_integer



from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from shutil import which
from subprocess import run
from typing import Optional

import pytest

import aristaproto
from tests.output_aristaproto import (
    map,
    nested,
    oneof,
    repeated,
    repeatedpacked,
)


oneof_example = oneof.Test().from_dict(
    {"pitied": 1, "just_a_regular_field": 123456789, "bar_name": "Testing"}
)

len_oneof = len(oneof_example)

nested_example = nested.Test().from_dict(
    {
        "nested": {"count": 1},
        "sibling": {"foo": 2},
        "sibling2": {"foo": 3},
        "msg": nested.TestMsg.THIS,
    }
)

repeated_example = repeated.Test().from_dict({"names": ["blah", "Blah2"]})

packed_example = repeatedpacked.Test().from_dict(
    {"counts": [1, 2, 3], "signed": [-1, 2, -3], "fixed": [1.2, -2.3, 3.4]}
)

map_example = map.Test().from_dict({"counts": {"blah": 1, "Blah2": 2}})

streams_path = Path("tests/streams/")

java = which("java")


def test_load_varint_too_long():
    with BytesIO(
        b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01"
    ) as stream, pytest.raises(ValueError):
        aristaproto.load_varint(stream)

    with BytesIO(b"\x80\x80\x80\x80\x80\x80\x80\x80\x80\x01") as stream:
        # This should not raise a ValueError, as it is within 64 bits
        aristaproto.load_varint(stream)


def test_load_varint_file():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        assert aristaproto.load_varint(stream) == (8, b"\x08")  # Single-byte varint
        stream.read(2)  # Skip until first multi-byte
        assert aristaproto.load_varint(stream) == (
            123456789,
            b"\x95\x9A\xEF\x3A",
        )  # Multi-byte varint


def test_load_varint_cutoff():
    with open(streams_path / "load_varint_cutoff.in", "rb") as stream:
        with pytest.raises(EOFError):
            aristaproto.load_varint(stream)

        stream.seek(1)
        with pytest.raises(EOFError):
            aristaproto.load_varint(stream)


def test_dump_varint_file(tmp_path):
    # Dump test varints to file
    with open(tmp_path / "dump_varint_file.out", "wb") as stream:
        aristaproto.dump_varint(8, stream)  # Single-byte varint
        aristaproto.dump_varint(123456789, stream)  # Multi-byte varint

    # Check that file contents are as expected
    with open(tmp_path / "dump_varint_file.out", "rb") as test_stream, open(
        streams_path / "message_dump_file_single.expected", "rb"
    ) as exp_stream:
        assert aristaproto.load_varint(test_stream) == aristaproto.load_varint(
            exp_stream
        )
        exp_stream.read(2)
        assert aristaproto.load_varint(test_stream) == aristaproto.load_varint(
            exp_stream
        )


def test_parse_fields():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        parsed_bytes = aristaproto.parse_fields(stream.read())

    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        parsed_stream = aristaproto.load_fields(stream)
        for field in parsed_bytes:
            assert field == next(parsed_stream)


def test_message_dump_file_single(tmp_path):
    # Write the message to the stream
    with open(tmp_path / "message_dump_file_single.out", "wb") as stream:
        oneof_example.dump(stream)

    # Check that the outputted file is exactly as expected
    with open(tmp_path / "message_dump_file_single.out", "rb") as test_stream, open(
        streams_path / "message_dump_file_single.expected", "rb"
    ) as exp_stream:
        assert test_stream.read() == exp_stream.read()


def test_message_dump_file_multiple(tmp_path):
    # Write the same Message twice and another, different message
    with open(tmp_path / "message_dump_file_multiple.out", "wb") as stream:
        oneof_example.dump(stream)
        oneof_example.dump(stream)
        nested_example.dump(stream)

    # Check that all three Messages were outputted to the file correctly
    with open(tmp_path / "message_dump_file_multiple.out", "rb") as test_stream, open(
        streams_path / "message_dump_file_multiple.expected", "rb"
    ) as exp_stream:
        assert test_stream.read() == exp_stream.read()


def test_message_dump_delimited(tmp_path):
    with open(tmp_path / "message_dump_delimited.out", "wb") as stream:
        oneof_example.dump(stream, aristaproto.SIZE_DELIMITED)
        oneof_example.dump(stream, aristaproto.SIZE_DELIMITED)
        nested_example.dump(stream, aristaproto.SIZE_DELIMITED)

    with open(tmp_path / "message_dump_delimited.out", "rb") as test_stream, open(
        streams_path / "delimited_messages.in", "rb"
    ) as exp_stream:
        assert test_stream.read() == exp_stream.read()


def test_message_len():
    assert len_oneof == len(bytes(oneof_example))
    assert len(nested_example) == len(bytes(nested_example))


def test_message_load_file_single():
    with open(streams_path / "message_dump_file_single.expected", "rb") as stream:
        assert oneof.Test().load(stream) == oneof_example
        stream.seek(0)
        assert oneof.Test().load(stream, len_oneof) == oneof_example


def test_message_load_file_multiple():
    with open(streams_path / "message_dump_file_multiple.expected", "rb") as stream:
        oneof_size = len_oneof
        assert oneof.Test().load(stream, oneof_size) == oneof_example
        assert oneof.Test().load(stream, oneof_size) == oneof_example
        assert nested.Test().load(stream) == nested_example
        assert stream.read(1) == b""


def test_message_load_too_small():
    with open(
        streams_path / "message_dump_file_single.expected", "rb"
    ) as stream, pytest.raises(ValueError):
        oneof.Test().load(stream, len_oneof - 1)


def test_message_load_delimited():
    with open(streams_path / "delimited_messages.in", "rb") as stream:
        assert oneof.Test().load(stream, aristaproto.SIZE_DELIMITED) == oneof_example
        assert oneof.Test().load(stream, aristaproto.SIZE_DELIMITED) == oneof_example
        assert nested.Test().load(stream, aristaproto.SIZE_DELIMITED) == nested_example
        assert stream.read(1) == b""


def test_message_load_too_large():
    with open(
        streams_path / "message_dump_file_single.expected", "rb"
    ) as stream, pytest.raises(ValueError):
        oneof.Test().load(stream, len_oneof + 1)


def test_message_len_optional_field():
    @dataclass
    class Request(aristaproto.Message):
        flag: Optional[bool] = aristaproto.message_field(1, wraps=aristaproto.TYPE_BOOL)

    assert len(Request()) == len(b"")
    assert len(Request(flag=True)) == len(b"\n\x02\x08\x01")
    assert len(Request(flag=False)) == len(b"\n\x00")


def test_message_len_repeated_field():
    assert len(repeated_example) == len(bytes(repeated_example))


def test_message_len_packed_field():
    assert len(packed_example) == len(bytes(packed_example))


def test_message_len_map_field():
    assert len(map_example) == len(bytes(map_example))


def test_message_len_empty_string():
    @dataclass
    class Empty(aristaproto.Message):
        string: str = aristaproto.string_field(1, "group")
        integer: int = aristaproto.int32_field(2, "group")

    empty = Empty().from_dict({"string": ""})
    assert len(empty) == len(bytes(empty))


def test_calculate_varint_size_negative():
    single_byte = -1
    multi_byte = -10000000
    edge = -(1 << 63)
    beyond = -(1 << 63) - 1
    before = -(1 << 63) + 1

    assert (
        aristaproto.size_varint(single_byte)
        == len(aristaproto.encode_varint(single_byte))
        == 10
    )
    assert (
        aristaproto.size_varint(multi_byte)
        == len(aristaproto.encode_varint(multi_byte))
        == 10
    )
    assert aristaproto.size_varint(edge) == len(aristaproto.encode_varint(edge)) == 10
    assert (
        aristaproto.size_varint(before) == len(aristaproto.encode_varint(before)) == 10
    )

    with pytest.raises(ValueError):
        aristaproto.size_varint(beyond)


def test_calculate_varint_size_positive():
    single_byte = 1
    multi_byte = 10000000

    assert aristaproto.size_varint(single_byte) == len(
        aristaproto.encode_varint(single_byte)
    )
    assert aristaproto.size_varint(multi_byte) == len(
        aristaproto.encode_varint(multi_byte)
    )


def test_dump_varint_negative(tmp_path):
    single_byte = -1
    multi_byte = -10000000
    edge = -(1 << 63)
    beyond = -(1 << 63) - 1
    before = -(1 << 63) + 1

    with open(tmp_path / "dump_varint_negative.out", "wb") as stream:
        aristaproto.dump_varint(single_byte, stream)
        aristaproto.dump_varint(multi_byte, stream)
        aristaproto.dump_varint(edge, stream)
        aristaproto.dump_varint(before, stream)

        with pytest.raises(ValueError):
            aristaproto.dump_varint(beyond, stream)

    with open(streams_path / "dump_varint_negative.expected", "rb") as exp_stream, open(
        tmp_path / "dump_varint_negative.out", "rb"
    ) as test_stream:
        assert test_stream.read() == exp_stream.read()


def test_dump_varint_positive(tmp_path):
    single_byte = 1
    multi_byte = 10000000

    with open(tmp_path / "dump_varint_positive.out", "wb") as stream:
        aristaproto.dump_varint(single_byte, stream)
        aristaproto.dump_varint(multi_byte, stream)

    with open(tmp_path / "dump_varint_positive.out", "rb") as test_stream, open(
        streams_path / "dump_varint_positive.expected", "rb"
    ) as exp_stream:
        assert test_stream.read() == exp_stream.read()


# Java compatibility tests


@pytest.fixture(scope="module")
def compile_jar():
    # Skip if not all required tools are present
    if java is None:
        pytest.skip("`java` command is absent and is required")
    mvn = which("mvn")
    if mvn is None:
        pytest.skip("Maven is absent and is required")

    # Compile the JAR
    proc_maven = run([mvn, "clean", "install", "-f", "tests/streams/java/pom.xml"])
    if proc_maven.returncode != 0:
        pytest.skip(
            "Maven compatibility-test.jar build failed (maybe Java version <11?)"
        )


jar = "tests/streams/java/target/compatibility-test.jar"


def run_jar(command: str, tmp_path):
    return run([java, "-jar", jar, command, tmp_path], check=True)


def run_java_single_varint(value: int, tmp_path) -> int:
    # Write single varint to file
    with open(tmp_path / "py_single_varint.out", "wb") as stream:
        aristaproto.dump_varint(value, stream)

    # Have Java read this varint and write it back
    run_jar("single_varint", tmp_path)

    # Read single varint from Java output file
    with open(tmp_path / "java_single_varint.out", "rb") as stream:
        returned = aristaproto.load_varint(stream)
        with pytest.raises(EOFError):
            aristaproto.load_varint(stream)

    return returned


def test_single_varint(compile_jar, tmp_path):
    single_byte = (1, b"\x01")
    multi_byte = (123456789, b"\x95\x9A\xEF\x3A")

    # Write a single-byte varint to a file and have Java read it back
    returned = run_java_single_varint(single_byte[0], tmp_path)
    assert returned == single_byte

    # Same for a multi-byte varint
    returned = run_java_single_varint(multi_byte[0], tmp_path)
    assert returned == multi_byte


def test_multiple_varints(compile_jar, tmp_path):
    single_byte = (1, b"\x01")
    multi_byte = (123456789, b"\x95\x9A\xEF\x3A")
    over32 = (3000000000, b"\x80\xBC\xC1\x96\x0B")

    # Write two varints to the same file
    with open(tmp_path / "py_multiple_varints.out", "wb") as stream:
        aristaproto.dump_varint(single_byte[0], stream)
        aristaproto.dump_varint(multi_byte[0], stream)
        aristaproto.dump_varint(over32[0], stream)

    # Have Java read these varints and write them back
    run_jar("multiple_varints", tmp_path)

    # Read varints from Java output file
    with open(tmp_path / "java_multiple_varints.out", "rb") as stream:
        returned_single = aristaproto.load_varint(stream)
        returned_multi = aristaproto.load_varint(stream)
        returned_over32 = aristaproto.load_varint(stream)
        with pytest.raises(EOFError):
            aristaproto.load_varint(stream)

    assert returned_single == single_byte
    assert returned_multi == multi_byte
    assert returned_over32 == over32


def test_single_message(compile_jar, tmp_path):
    # Write message to file
    with open(tmp_path / "py_single_message.out", "wb") as stream:
        oneof_example.dump(stream)

    # Have Java read and return the message
    run_jar("single_message", tmp_path)

    # Read and check the returned message
    with open(tmp_path / "java_single_message.out", "rb") as stream:
        returned = oneof.Test().load(stream, len(bytes(oneof_example)))
        assert stream.read() == b""

    assert returned == oneof_example


def test_multiple_messages(compile_jar, tmp_path):
    # Write delimited messages to file
    with open(tmp_path / "py_multiple_messages.out", "wb") as stream:
        oneof_example.dump(stream, aristaproto.SIZE_DELIMITED)
        nested_example.dump(stream, aristaproto.SIZE_DELIMITED)

    # Have Java read and return the messages
    run_jar("multiple_messages", tmp_path)

    # Read and check the returned messages
    with open(tmp_path / "java_multiple_messages.out", "rb") as stream:
        returned_oneof = oneof.Test().load(stream, aristaproto.SIZE_DELIMITED)
        returned_nested = nested.Test().load(stream, aristaproto.SIZE_DELIMITED)
        assert stream.read() == b""

    assert returned_oneof == oneof_example
    assert returned_nested == nested_example


def test_infinite_messages(compile_jar, tmp_path):
    num_messages = 5

    # Write delimited messages to file
    with open(tmp_path / "py_infinite_messages.out", "wb") as stream:
        for x in range(num_messages):
            oneof_example.dump(stream, aristaproto.SIZE_DELIMITED)

    # Have Java read and return the messages
    run_jar("infinite_messages", tmp_path)

    # Read and check the returned messages
    messages = []
    with open(tmp_path / "java_infinite_messages.out", "rb") as stream:
        while True:
            try:
                messages.append(oneof.Test().load(stream, aristaproto.SIZE_DELIMITED))
            except EOFError:
                break

    assert len(messages) == num_messages



from datetime import (
    datetime,
    timezone,
)

import pytest
from google.protobuf import json_format
from google.protobuf.timestamp_pb2 import Timestamp

import aristaproto
from tests.output_aristaproto.google_impl_behavior_equivalence import (
    Empty,
    Foo,
    Request,
    Spam,
    Test,
)
from tests.output_reference.google_impl_behavior_equivalence.google_impl_behavior_equivalence_pb2 import (
    Empty as ReferenceEmpty,
    Foo as ReferenceFoo,
    Request as ReferenceRequest,
    Spam as ReferenceSpam,
    Test as ReferenceTest,
)


def test_oneof_serializes_similar_to_google_oneof():
    tests = [
        (Test(string="abc"), ReferenceTest(string="abc")),
        (Test(integer=2), ReferenceTest(integer=2)),
        (Test(foo=Foo(bar=1)), ReferenceTest(foo=ReferenceFoo(bar=1))),
        # Default values should also behave the same within oneofs
        (Test(string=""), ReferenceTest(string="")),
        (Test(integer=0), ReferenceTest(integer=0)),
        (Test(foo=Foo(bar=0)), ReferenceTest(foo=ReferenceFoo(bar=0))),
    ]
    for message, message_reference in tests:
        # NOTE: As of July 2020, MessageToJson inserts newlines in the output string so,
        # just compare dicts
        assert message.to_dict() == json_format.MessageToDict(message_reference)


def test_bytes_are_the_same_for_oneof():
    message = Test(string="")
    message_reference = ReferenceTest(string="")

    message_bytes = bytes(message)
    message_reference_bytes = message_reference.SerializeToString()

    assert message_bytes == message_reference_bytes

    message2 = Test().parse(message_reference_bytes)
    message_reference2 = ReferenceTest()
    message_reference2.ParseFromString(message_reference_bytes)

    assert message == message2
    assert message_reference == message_reference2

    # None of these fields were explicitly set BUT they should not actually be null
    # themselves
    assert not hasattr(message, "foo")
    assert object.__getattribute__(message, "foo") == aristaproto.PLACEHOLDER
    assert not hasattr(message2, "foo")
    assert object.__getattribute__(message2, "foo") == aristaproto.PLACEHOLDER

    assert isinstance(message_reference.foo, ReferenceFoo)
    assert isinstance(message_reference2.foo, ReferenceFoo)


@pytest.mark.parametrize("dt", (datetime.min.replace(tzinfo=timezone.utc),))
def test_datetime_clamping(dt):  # see #407
    ts = Timestamp()
    ts.FromDatetime(dt)
    assert bytes(Spam(dt)) == ReferenceSpam(ts=ts).SerializeToString()
    message_bytes = bytes(Spam(dt))

    assert (
        Spam().parse(message_bytes).ts.timestamp()
        == ReferenceSpam.FromString(message_bytes).ts.seconds
    )


def test_empty_message_field():
    message = Request()
    reference_message = ReferenceRequest()

    message.foo = Empty()
    reference_message.foo.CopyFrom(ReferenceEmpty())

    assert aristaproto.serialized_on_wire(message.foo)
    assert reference_message.HasField("foo")

    assert bytes(message) == reference_message.SerializeToString()



from tests.output_aristaproto.proto3_field_presence_oneof import (
    InnerNested,
    Nested,
    Test,
    WithOptional,
)


def test_serialization():
    """Ensure that serialization of fields unset but with explicit field
    presence do not bloat the serialized payload with length-delimited fields
    with length 0"""

    def test_empty_nested(message: Test) -> None:
        # '0a' => tag 1, length delimited
        # '00' => length: 0
        assert bytes(message) == bytearray.fromhex("0a 00")

    test_empty_nested(Test(nested=Nested()))
    test_empty_nested(Test(nested=Nested(inner=None)))
    test_empty_nested(Test(nested=Nested(inner=InnerNested(a=None))))

    def test_empty_with_optional(message: Test) -> None:
        # '12' => tag 2, length delimited
        # '00' => length: 0
        assert bytes(message) == bytearray.fromhex("12 00")

    test_empty_with_optional(Test(with_optional=WithOptional()))
    test_empty_with_optional(Test(with_optional=WithOptional(b=None)))



from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Any,
    Callable,
)

import pytest

import aristaproto.lib.google.protobuf as protobuf
from tests.mocks import MockChannel
from tests.output_aristaproto.googletypes_request import (
    Input,
    TestStub,
)


test_cases = [
    (TestStub.send_double, protobuf.DoubleValue, 2.5),
    (TestStub.send_float, protobuf.FloatValue, 2.5),
    (TestStub.send_int64, protobuf.Int64Value, -64),
    (TestStub.send_u_int64, protobuf.UInt64Value, 64),
    (TestStub.send_int32, protobuf.Int32Value, -32),
    (TestStub.send_u_int32, protobuf.UInt32Value, 32),
    (TestStub.send_bool, protobuf.BoolValue, True),
    (TestStub.send_string, protobuf.StringValue, "string"),
    (TestStub.send_bytes, protobuf.BytesValue, bytes(0xFF)[0:4]),
    (TestStub.send_datetime, protobuf.Timestamp, datetime(2038, 1, 19, 3, 14, 8)),
    (TestStub.send_timedelta, protobuf.Duration, timedelta(seconds=123456)),
]


@pytest.mark.asyncio
@pytest.mark.parametrize(["service_method", "wrapper_class", "value"], test_cases)
async def test_channel_receives_wrapped_type(
    service_method: Callable[[TestStub, Input], Any], wrapper_class: Callable, value
):
    wrapped_value = wrapper_class()
    wrapped_value.value = value
    channel = MockChannel(responses=[Input()])
    service = TestStub(channel)

    await service_method(service, wrapped_value)

    assert channel.requests[0]["request"] == type(wrapped_value)



import pytest

from tests.mocks import MockChannel
from tests.output_aristaproto.import_service_input_message import (
    NestedRequestMessage,
    RequestMessage,
    RequestResponse,
    TestStub,
)
from tests.output_aristaproto.import_service_input_message.child import (
    ChildRequestMessage,
)


@pytest.mark.asyncio
async def test_service_correctly_imports_reference_message():
    mock_response = RequestResponse(value=10)
    service = TestStub(MockChannel([mock_response]))
    response = await service.do_thing(RequestMessage(1))
    assert mock_response == response


@pytest.mark.asyncio
async def test_service_correctly_imports_reference_message_from_child_package():
    mock_response = RequestResponse(value=10)
    service = TestStub(MockChannel([mock_response]))
    response = await service.do_thing2(ChildRequestMessage(1))
    assert mock_response == response


@pytest.mark.asyncio
async def test_service_correctly_imports_nested_reference():
    mock_response = RequestResponse(value=10)
    service = TestStub(MockChannel([mock_response]))
    response = await service.do_thing3(NestedRequestMessage(1))
    assert mock_response == response



import keyword
import re


# Word delimiters and symbols that will not be preserved when re-casing.
# language=PythonRegExp
SYMBOLS = "[^a-zA-Z0-9]*"

# Optionally capitalized word.
# language=PythonRegExp
WORD = "[A-Z]*[a-z]*[0-9]*"

# Uppercase word, not followed by lowercase letters.
# language=PythonRegExp
WORD_UPPER = "[A-Z]+(?![a-z])[0-9]*"


def safe_snake_case(value: str) -> str:
    """Snake case a value taking into account Python keywords."""
    value = snake_case(value)
    value = sanitize_name(value)
    return value


def snake_case(value: str, strict: bool = True) -> str:
    """
    Join words with an underscore into lowercase and remove symbols.

    Parameters
    -----------
    value: :class:`str`
        The value to convert.
    strict: :class:`bool`
        Whether or not to force single underscores.

    Returns
    --------
    :class:`str`
        The value in snake_case.
    """

    def substitute_word(symbols: str, word: str, is_start: bool) -> str:
        if not word:
            return ""
        if strict:
            delimiter_count = 0 if is_start else 1  # Single underscore if strict.
        elif is_start:
            delimiter_count = len(symbols)
        elif word.isupper() or word.islower():
            delimiter_count = max(
                1, len(symbols)
            )  # Preserve all delimiters if not strict.
        else:
            delimiter_count = len(symbols) + 1  # Extra underscore for leading capital.

        return ("_" * delimiter_count) + word.lower()

    snake = re.sub(
        f"(^)?({SYMBOLS})({WORD_UPPER}|{WORD})",
        lambda groups: substitute_word(groups[2], groups[3], groups[1] is not None),
        value,
    )
    return snake


def pascal_case(value: str, strict: bool = True) -> str:
    """
    Capitalize each word and remove symbols.

    Parameters
    -----------
    value: :class:`str`
        The value to convert.
    strict: :class:`bool`
        Whether or not to output only alphanumeric characters.

    Returns
    --------
    :class:`str`
        The value in PascalCase.
    """

    def substitute_word(symbols, word):
        if strict:
            return word.capitalize()  # Remove all delimiters

        if word.islower():
            delimiter_length = len(symbols[:-1])  # Lose one delimiter
        else:
            delimiter_length = len(symbols)  # Preserve all delimiters

        return ("_" * delimiter_length) + word.capitalize()

    return re.sub(
        f"({SYMBOLS})({WORD_UPPER}|{WORD})",
        lambda groups: substitute_word(groups[1], groups[2]),
        value,
    )


def camel_case(value: str, strict: bool = True) -> str:
    """
    Capitalize all words except first and remove symbols.

    Parameters
    -----------
    value: :class:`str`
        The value to convert.
    strict: :class:`bool`
        Whether or not to output only alphanumeric characters.

    Returns
    --------
    :class:`str`
        The value in camelCase.
    """
    return lowercase_first(pascal_case(value, strict=strict))


def lowercase_first(value: str) -> str:
    """
    Lower cases the first character of the value.

    Parameters
    ----------
    value: :class:`str`
        The value to lower case.

    Returns
    -------
    :class:`str`
        The lower cased string.
    """
    return value[0:1].lower() + value[1:]


def sanitize_name(value: str) -> str:
    # https://www.python.org/dev/peps/pep-0008/#descriptive-naming-styles
    if keyword.iskeyword(value):
        return f"{value}_"
    if not value.isidentifier():
        return f"_{value}"
    return value



import json

from aristaproto.lib.google.protobuf import Struct
from aristaproto.lib.pydantic.google.protobuf import Struct as StructPydantic


def test_struct_roundtrip():
    data = {
        "foo": "bar",
        "baz": None,
        "quux": 123,
        "zap": [1, {"two": 3}, "four"],
    }
    data_json = json.dumps(data)

    struct_from_dict = Struct().from_dict(data)
    assert struct_from_dict.fields == data
    assert struct_from_dict.to_dict() == data
    assert struct_from_dict.to_json() == data_json

    struct_from_json = Struct().from_json(data_json)
    assert struct_from_json.fields == data
    assert struct_from_json.to_dict() == data
    assert struct_from_json == struct_from_dict
    assert struct_from_json.to_json() == data_json

    struct_pyd_from_dict = StructPydantic(fields={}).from_dict(data)
    assert struct_pyd_from_dict.fields == data
    assert struct_pyd_from_dict.to_dict() == data
    assert struct_pyd_from_dict.to_json() == data_json

    struct_pyd_from_dict = StructPydantic(fields={}).from_json(data_json)
    assert struct_pyd_from_dict.fields == data
    assert struct_pyd_from_dict.to_dict() == data
    assert struct_pyd_from_dict == struct_pyd_from_dict
    assert struct_pyd_from_dict.to_json() == data_json



import os.path


try:
    # aristaproto[compiler] specific dependencies
    import black
    import isort.api
    import jinja2
except ImportError as err:
    print(
        "\033[31m"
        f"Unable to import `{err.name}` from aristaproto plugin! "
        "Please ensure that you've installed aristaproto as "
        '`pip install "aristaproto[compiler]"` so that compiler dependencies '
        "are included."
        "\033[0m"
    )
    raise SystemExit(1)

from .models import OutputTemplate


def outputfile_compiler(output_file: OutputTemplate) -> str:
    templates_folder = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "templates")
    )

    env = jinja2.Environment(
        trim_blocks=True,
        lstrip_blocks=True,
        loader=jinja2.FileSystemLoader(templates_folder),
    )
    template = env.get_template("template.py.j2")

    code = template.render(output_file=output_file)
    code = isort.api.sort_code_string(
        code=code,
        show_diff=False,
        py_version=37,
        profile="black",
        combine_as_imports=True,
        lines_after_imports=2,
        quiet=True,
        force_grid_wrap=2,
        known_third_party=["grpclib", "aristaproto"],
    )
    return black.format_str(
        src_contents=code,
        mode=black.Mode(),
    )



from datetime import (
    datetime,
    timedelta,
    timezone,
)

import pytest

from tests.output_aristaproto.timestamp_dict_encode import Test


# Current World Timezone range (UTC-12 to UTC+14)
MIN_UTC_OFFSET_MIN = -12 * 60
MAX_UTC_OFFSET_MIN = 14 * 60

# Generate all timezones in range in 15 min increments
timezones = [
    timezone(timedelta(minutes=x))
    for x in range(MIN_UTC_OFFSET_MIN, MAX_UTC_OFFSET_MIN + 1, 15)
]


@pytest.mark.parametrize("tz", timezones)
def test_timezone_aware_datetime_dict_encode(tz: timezone):
    original_time = datetime.now(tz=tz)
    original_message = Test()
    original_message.ts = original_time
    encoded = original_message.to_dict()
    decoded_message = Test()
    decoded_message.from_dict(encoded)

    # check that the timestamps are equal after decoding from dict
    assert original_message.ts.tzinfo is not None
    assert decoded_message.ts.tzinfo is not None
    assert original_message.ts == decoded_message.ts


def test_naive_datetime_dict_encode():
    # make suer naive datetime objects are still treated as utc
    original_time = datetime.now()
    assert original_time.tzinfo is None
    original_message = Test()
    original_message.ts = original_time
    original_time_utc = original_time.replace(tzinfo=timezone.utc)
    encoded = original_message.to_dict()
    decoded_message = Test()
    decoded_message.from_dict(encoded)

    # check that the timestamps are equal after decoding from dict
    assert decoded_message.ts.tzinfo is not None
    assert original_time_utc == decoded_message.ts


@pytest.mark.parametrize("tz", timezones)
def test_timezone_aware_json_serialize(tz: timezone):
    original_time = datetime.now(tz=tz)
    original_message = Test()
    original_message.ts = original_time
    json_serialized = original_message.to_json()
    decoded_message = Test()
    decoded_message.from_json(json_serialized)

    # check that the timestamps are equal after decoding from dict
    assert original_message.ts.tzinfo is not None
    assert decoded_message.ts.tzinfo is not None
    assert original_message.ts == decoded_message.ts


def test_naive_datetime_json_serialize():
    # make suer naive datetime objects are still treated as utc
    original_time = datetime.now()
    assert original_time.tzinfo is None
    original_message = Test()
    original_message.ts = original_time
    original_time_utc = original_time.replace(tzinfo=timezone.utc)
    json_serialized = original_message.to_json()
    decoded_message = Test()
    decoded_message.from_json(json_serialized)

    # check that the timestamps are equal after decoding from dict
    assert decoded_message.ts.tzinfo is not None
    assert original_time_utc == decoded_message.ts



import importlib
import json
import math
import os
import sys
from collections import namedtuple
from types import ModuleType
from typing import (
    Any,
    Dict,
    List,
    Set,
    Tuple,
)

import pytest

import aristaproto
from tests.inputs import config as test_input_config
from tests.mocks import MockChannel
from tests.util import (
    find_module,
    get_directories,
    get_test_case_json_data,
    inputs_path,
)


# Force pure-python implementation instead of C++, otherwise imports
# break things because we can't properly reset the symbol database.
os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

from google.protobuf.json_format import Parse


class TestCases:
    def __init__(
        self,
        path,
        services: Set[str],
        xfail: Set[str],
    ):
        _all = set(get_directories(path)) - {"__pycache__"}
        _services = services
        _messages = (_all - services) - {"__pycache__"}
        _messages_with_json = {
            test for test in _messages if get_test_case_json_data(test)
        }

        unknown_xfail_tests = xfail - _all
        if unknown_xfail_tests:
            raise Exception(f"Unknown test(s) in config.py: {unknown_xfail_tests}")

        self.all = self.apply_xfail_marks(_all, xfail)
        self.services = self.apply_xfail_marks(_services, xfail)
        self.messages = self.apply_xfail_marks(_messages, xfail)
        self.messages_with_json = self.apply_xfail_marks(_messages_with_json, xfail)

    @staticmethod
    def apply_xfail_marks(test_set: Set[str], xfail: Set[str]):
        return [
            pytest.param(test, marks=pytest.mark.xfail) if test in xfail else test
            for test in test_set
        ]


test_cases = TestCases(
    path=inputs_path,
    services=test_input_config.services,
    xfail=test_input_config.xfail,
)

plugin_output_package = "tests.output_aristaproto"
reference_output_package = "tests.output_reference"

TestData = namedtuple("TestData", ["plugin_module", "reference_module", "json_data"])


def module_has_entry_point(module: ModuleType):
    return any(hasattr(module, attr) for attr in ["Test", "TestStub"])


def list_replace_nans(items: List) -> List[Any]:
    """Replace float("nan") in a list with the string "NaN"

    Parameters
    ----------
    items : List
            List to update

    Returns
    -------
    List[Any]
        Updated list
    """
    result = []
    for item in items:
        if isinstance(item, list):
            result.append(list_replace_nans(item))
        elif isinstance(item, dict):
            result.append(dict_replace_nans(item))
        elif isinstance(item, float) and math.isnan(item):
            result.append(aristaproto.NAN)
    return result


def dict_replace_nans(input_dict: Dict[Any, Any]) -> Dict[Any, Any]:
    """Replace float("nan") in a dictionary with the string "NaN"

    Parameters
    ----------
    input_dict : Dict[Any, Any]
            Dictionary to update

    Returns
    -------
    Dict[Any, Any]
        Updated dictionary
    """
    result = {}
    for key, value in input_dict.items():
        if isinstance(value, dict):
            value = dict_replace_nans(value)
        elif isinstance(value, list):
            value = list_replace_nans(value)
        elif isinstance(value, float) and math.isnan(value):
            value = aristaproto.NAN
        result[key] = value
    return result


@pytest.fixture
def test_data(request, reset_sys_path):
    test_case_name = request.param

    reference_module_root = os.path.join(
        *reference_output_package.split("."), test_case_name
    )
    sys.path.append(reference_module_root)

    plugin_module = importlib.import_module(f"{plugin_output_package}.{test_case_name}")

    plugin_module_entry_point = find_module(plugin_module, module_has_entry_point)

    if not plugin_module_entry_point:
        raise Exception(
            f"Test case {repr(test_case_name)} has no entry point. "
            "Please add a proto message or service called Test and recompile."
        )

    yield (
        TestData(
            plugin_module=plugin_module_entry_point,
            reference_module=lambda: importlib.import_module(
                f"{reference_output_package}.{test_case_name}.{test_case_name}_pb2"
            ),
            json_data=get_test_case_json_data(test_case_name),
        )
    )


@pytest.mark.parametrize("test_data", test_cases.messages, indirect=True)
def test_message_can_instantiated(test_data: TestData) -> None:
    plugin_module, *_ = test_data
    plugin_module.Test()


@pytest.mark.parametrize("test_data", test_cases.messages, indirect=True)
def test_message_equality(test_data: TestData) -> None:
    plugin_module, *_ = test_data
    message1 = plugin_module.Test()
    message2 = plugin_module.Test()
    assert message1 == message2


@pytest.mark.parametrize("test_data", test_cases.messages_with_json, indirect=True)
def test_message_json(repeat, test_data: TestData) -> None:
    plugin_module, _, json_data = test_data

    for _ in range(repeat):
        for sample in json_data:
            if sample.belongs_to(test_input_config.non_symmetrical_json):
                continue

            message: aristaproto.Message = plugin_module.Test()

            message.from_json(sample.json)
            message_json = message.to_json(0)

            assert dict_replace_nans(json.loads(message_json)) == dict_replace_nans(
                json.loads(sample.json)
            )


@pytest.mark.parametrize("test_data", test_cases.services, indirect=True)
def test_service_can_be_instantiated(test_data: TestData) -> None:
    test_data.plugin_module.TestStub(MockChannel())


@pytest.mark.parametrize("test_data", test_cases.messages_with_json, indirect=True)
def test_binary_compatibility(repeat, test_data: TestData) -> None:
    plugin_module, reference_module, json_data = test_data

    for sample in json_data:
        reference_instance = Parse(sample.json, reference_module().Test())
        reference_binary_output = reference_instance.SerializeToString()

        for _ in range(repeat):
            plugin_instance_from_json: aristaproto.Message = (
                plugin_module.Test().from_json(sample.json)
            )
            plugin_instance_from_binary = plugin_module.Test.FromString(
                reference_binary_output
            )

            # Generally this can't be relied on, but here we are aiming to match the
            # existing Python implementation and aren't doing anything tricky.
            # https://developers.google.com/protocol-buffers/docs/encoding#implications
            assert bytes(plugin_instance_from_json) == reference_binary_output
            assert bytes(plugin_instance_from_binary) == reference_binary_output

            assert plugin_instance_from_json == plugin_instance_from_binary
            assert dict_replace_nans(
                plugin_instance_from_json.to_dict()
            ) == dict_replace_nans(plugin_instance_from_binary.to_dict())



import tests.output_aristaproto.casing_inner_class as casing_inner_class


def test_message_casing_inner_class_name():
    assert hasattr(
        casing_inner_class, "TestInnerClass"
    ), "Inline defined Message is correctly converted to CamelCase"


def test_message_casing_inner_class_attributes():
    message = casing_inner_class.Test()
    assert hasattr(
        message.inner, "old_exp"
    ), "Inline defined Message attribute is snake_case"



import pytest

from tests.output_aristaproto.bool import Test
from tests.output_aristaproto_pydantic.bool import Test as TestPyd


def test_value():
    message = Test()
    assert not message.value, "Boolean is False by default"


def test_pydantic_no_value():
    with pytest.raises(ValueError):
        TestPyd()


def test_pydantic_value():
    message = Test(value=False)
    assert not message.value



# Test cases that are expected to fail, e.g. unimplemented features or bug-fixes.
# Remove from list when fixed.
xfail = {
    "namespace_keywords",  # 70
    "googletypes_struct",  # 9
    "googletypes_value",  # 9
    "import_capitalized_package",
    "example",  # This is the example in the readme. Not a test.
}

services = {
    "googletypes_request",
    "googletypes_response",
    "googletypes_response_embedded",
    "service",
    "service_separate_packages",
    "import_service_input_message",
    "googletypes_service_returns_empty",
    "googletypes_service_returns_googletype",
    "example_service",
    "empty_service",
    "service_uppercase",
}


# Indicate json sample messages to skip when testing that json (de)serialization
# is symmetrical becuase some cases legitimately are not symmetrical.
# Each key references the name of the test scenario and the values in the tuple
# Are the names of the json files.
non_symmetrical_json = {"empty_repeated": ("empty_repeated",)}



import pytest

from aristaproto.casing import (
    camel_case,
    pascal_case,
    snake_case,
)


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("", ""),
        ("a", "A"),
        ("foobar", "Foobar"),
        ("fooBar", "FooBar"),
        ("FooBar", "FooBar"),
        ("foo.bar", "FooBar"),
        ("foo_bar", "FooBar"),
        ("FOOBAR", "Foobar"),
        ("FOOBar", "FooBar"),
        ("UInt32", "UInt32"),
        ("FOO_BAR", "FooBar"),
        ("FOOBAR1", "Foobar1"),
        ("FOOBAR_1", "Foobar1"),
        ("FOO1BAR2", "Foo1Bar2"),
        ("foo__bar", "FooBar"),
        ("_foobar", "Foobar"),
        ("foobaR", "FoobaR"),
        ("foo~bar", "FooBar"),
        ("foo:bar", "FooBar"),
        ("1foobar", "1Foobar"),
    ],
)
def test_pascal_case(value, expected):
    actual = pascal_case(value, strict=True)
    assert actual == expected, f"{value} => {expected} (actual: {actual})"


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("", ""),
        ("a", "a"),
        ("foobar", "foobar"),
        ("fooBar", "fooBar"),
        ("FooBar", "fooBar"),
        ("foo.bar", "fooBar"),
        ("foo_bar", "fooBar"),
        ("FOOBAR", "foobar"),
        ("FOO_BAR", "fooBar"),
        ("FOOBAR1", "foobar1"),
        ("FOOBAR_1", "foobar1"),
        ("FOO1BAR2", "foo1Bar2"),
        ("foo__bar", "fooBar"),
        ("_foobar", "foobar"),
        ("foobaR", "foobaR"),
        ("foo~bar", "fooBar"),
        ("foo:bar", "fooBar"),
        ("1foobar", "1Foobar"),
    ],
)
def test_camel_case_strict(value, expected):
    actual = camel_case(value, strict=True)
    assert actual == expected, f"{value} => {expected} (actual: {actual})"


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("foo_bar", "fooBar"),
        ("FooBar", "fooBar"),
        ("foo__bar", "foo_Bar"),
        ("foo__Bar", "foo__Bar"),
    ],
)
def test_camel_case_not_strict(value, expected):
    actual = camel_case(value, strict=False)
    assert actual == expected, f"{value} => {expected} (actual: {actual})"


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("", ""),
        ("a", "a"),
        ("foobar", "foobar"),
        ("fooBar", "foo_bar"),
        ("FooBar", "foo_bar"),
        ("foo.bar", "foo_bar"),
        ("foo_bar", "foo_bar"),
        ("foo_Bar", "foo_bar"),
        ("FOOBAR", "foobar"),
        ("FOOBar", "foo_bar"),
        ("UInt32", "u_int32"),
        ("FOO_BAR", "foo_bar"),
        ("FOOBAR1", "foobar1"),
        ("FOOBAR_1", "foobar_1"),
        ("FOOBAR_123", "foobar_123"),
        ("FOO1BAR2", "foo1_bar2"),
        ("foo__bar", "foo_bar"),
        ("_foobar", "foobar"),
        ("foobaR", "fooba_r"),
        ("foo~bar", "foo_bar"),
        ("foo:bar", "foo_bar"),
        ("1foobar", "1_foobar"),
        ("GetUInt64", "get_u_int64"),
    ],
)
def test_snake_case_strict(value, expected):
    actual = snake_case(value)
    assert actual == expected, f"{value} => {expected} (actual: {actual})"


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("fooBar", "foo_bar"),
        ("FooBar", "foo_bar"),
        ("foo_Bar", "foo__bar"),
        ("foo__bar", "foo__bar"),
        ("FOOBar", "foo_bar"),
        ("__foo", "__foo"),
        ("GetUInt64", "get_u_int64"),
    ],
)
def test_snake_case_not_strict(value, expected):
    actual = snake_case(value, strict=False)
    assert actual == expected, f"{value} => {expected} (actual: {actual})"



from tests.output_aristaproto.regression_414 import Test


def test_full_cycle():
    body = bytes([0, 1])
    auth = bytes([2, 3])
    sig = [b""]

    obj = Test(body=body, auth=auth, signatures=sig)

    decoded = Test().parse(bytes(obj))
    assert decoded == obj
    assert decoded.body == body
    assert decoded.auth == auth
    assert decoded.signatures == sig



import pytest

from aristaproto.compile.importing import (
    get_type_reference,
    parse_source_type_name,
)


@pytest.mark.parametrize(
    ["google_type", "expected_name", "expected_import"],
    [
        (
            ".google.protobuf.Empty",
            '"aristaproto_lib_google_protobuf.Empty"',
            "import aristaproto.lib.google.protobuf as aristaproto_lib_google_protobuf",
        ),
        (
            ".google.protobuf.Struct",
            '"aristaproto_lib_google_protobuf.Struct"',
            "import aristaproto.lib.google.protobuf as aristaproto_lib_google_protobuf",
        ),
        (
            ".google.protobuf.ListValue",
            '"aristaproto_lib_google_protobuf.ListValue"',
            "import aristaproto.lib.google.protobuf as aristaproto_lib_google_protobuf",
        ),
        (
            ".google.protobuf.Value",
            '"aristaproto_lib_google_protobuf.Value"',
            "import aristaproto.lib.google.protobuf as aristaproto_lib_google_protobuf",
        ),
    ],
)
def test_reference_google_wellknown_types_non_wrappers(
    google_type: str, expected_name: str, expected_import: str
):
    imports = set()
    name = get_type_reference(
        package="", imports=imports, source_type=google_type, pydantic=False
    )

    assert name == expected_name
    assert imports.__contains__(
        expected_import
    ), f"{expected_import} not found in {imports}"


@pytest.mark.parametrize(
    ["google_type", "expected_name", "expected_import"],
    [
        (
            ".google.protobuf.Empty",
            '"aristaproto_lib_pydantic_google_protobuf.Empty"',
            "import aristaproto.lib.pydantic.google.protobuf as aristaproto_lib_pydantic_google_protobuf",
        ),
        (
            ".google.protobuf.Struct",
            '"aristaproto_lib_pydantic_google_protobuf.Struct"',
            "import aristaproto.lib.pydantic.google.protobuf as aristaproto_lib_pydantic_google_protobuf",
        ),
        (
            ".google.protobuf.ListValue",
            '"aristaproto_lib_pydantic_google_protobuf.ListValue"',
            "import aristaproto.lib.pydantic.google.protobuf as aristaproto_lib_pydantic_google_protobuf",
        ),
        (
            ".google.protobuf.Value",
            '"aristaproto_lib_pydantic_google_protobuf.Value"',
            "import aristaproto.lib.pydantic.google.protobuf as aristaproto_lib_pydantic_google_protobuf",
        ),
    ],
)
def test_reference_google_wellknown_types_non_wrappers_pydantic(
    google_type: str, expected_name: str, expected_import: str
):
    imports = set()
    name = get_type_reference(
        package="", imports=imports, source_type=google_type, pydantic=True
    )

    assert name == expected_name
    assert imports.__contains__(
        expected_import
    ), f"{expected_import} not found in {imports}"


@pytest.mark.parametrize(
    ["google_type", "expected_name"],
    [
        (".google.protobuf.DoubleValue", "Optional[float]"),
        (".google.protobuf.FloatValue", "Optional[float]"),
        (".google.protobuf.Int32Value", "Optional[int]"),
        (".google.protobuf.Int64Value", "Optional[int]"),
        (".google.protobuf.UInt32Value", "Optional[int]"),
        (".google.protobuf.UInt64Value", "Optional[int]"),
        (".google.protobuf.BoolValue", "Optional[bool]"),
        (".google.protobuf.StringValue", "Optional[str]"),
        (".google.protobuf.BytesValue", "Optional[bytes]"),
    ],
)
def test_referenceing_google_wrappers_unwraps_them(
    google_type: str, expected_name: str
):
    imports = set()
    name = get_type_reference(package="", imports=imports, source_type=google_type)

    assert name == expected_name
    assert imports == set()


@pytest.mark.parametrize(
    ["google_type", "expected_name"],
    [
        (
            ".google.protobuf.DoubleValue",
            '"aristaproto_lib_google_protobuf.DoubleValue"',
        ),
        (".google.protobuf.FloatValue", '"aristaproto_lib_google_protobuf.FloatValue"'),
        (".google.protobuf.Int32Value", '"aristaproto_lib_google_protobuf.Int32Value"'),
        (".google.protobuf.Int64Value", '"aristaproto_lib_google_protobuf.Int64Value"'),
        (
            ".google.protobuf.UInt32Value",
            '"aristaproto_lib_google_protobuf.UInt32Value"',
        ),
        (
            ".google.protobuf.UInt64Value",
            '"aristaproto_lib_google_protobuf.UInt64Value"',
        ),
        (".google.protobuf.BoolValue", '"aristaproto_lib_google_protobuf.BoolValue"'),
        (
            ".google.protobuf.StringValue",
            '"aristaproto_lib_google_protobuf.StringValue"',
        ),
        (".google.protobuf.BytesValue", '"aristaproto_lib_google_protobuf.BytesValue"'),
    ],
)
def test_referenceing_google_wrappers_without_unwrapping(
    google_type: str, expected_name: str
):
    name = get_type_reference(
        package="", imports=set(), source_type=google_type, unwrap=False
    )

    assert name == expected_name


def test_reference_child_package_from_package():
    imports = set()
    name = get_type_reference(
        package="package", imports=imports, source_type="package.child.Message"
    )

    assert imports == {"from . import child"}
    assert name == '"child.Message"'


def test_reference_child_package_from_root():
    imports = set()
    name = get_type_reference(package="", imports=imports, source_type="child.Message")

    assert imports == {"from . import child"}
    assert name == '"child.Message"'


def test_reference_camel_cased():
    imports = set()
    name = get_type_reference(
        package="", imports=imports, source_type="child_package.example_message"
    )

    assert imports == {"from . import child_package"}
    assert name == '"child_package.ExampleMessage"'


def test_reference_nested_child_from_root():
    imports = set()
    name = get_type_reference(
        package="", imports=imports, source_type="nested.child.Message"
    )

    assert imports == {"from .nested import child as nested_child"}
    assert name == '"nested_child.Message"'


def test_reference_deeply_nested_child_from_root():
    imports = set()
    name = get_type_reference(
        package="", imports=imports, source_type="deeply.nested.child.Message"
    )

    assert imports == {"from .deeply.nested import child as deeply_nested_child"}
    assert name == '"deeply_nested_child.Message"'


def test_reference_deeply_nested_child_from_package():
    imports = set()
    name = get_type_reference(
        package="package",
        imports=imports,
        source_type="package.deeply.nested.child.Message",
    )

    assert imports == {"from .deeply.nested import child as deeply_nested_child"}
    assert name == '"deeply_nested_child.Message"'


def test_reference_root_sibling():
    imports = set()
    name = get_type_reference(package="", imports=imports, source_type="Message")

    assert imports == set()
    assert name == '"Message"'


def test_reference_nested_siblings():
    imports = set()
    name = get_type_reference(package="foo", imports=imports, source_type="foo.Message")

    assert imports == set()
    assert name == '"Message"'


def test_reference_deeply_nested_siblings():
    imports = set()
    name = get_type_reference(
        package="foo.bar", imports=imports, source_type="foo.bar.Message"
    )

    assert imports == set()
    assert name == '"Message"'


def test_reference_parent_package_from_child():
    imports = set()
    name = get_type_reference(
        package="package.child", imports=imports, source_type="package.Message"
    )

    assert imports == {"from ... import package as __package__"}
    assert name == '"__package__.Message"'


def test_reference_parent_package_from_deeply_nested_child():
    imports = set()
    name = get_type_reference(
        package="package.deeply.nested.child",
        imports=imports,
        source_type="package.deeply.nested.Message",
    )

    assert imports == {"from ... import nested as __nested__"}
    assert name == '"__nested__.Message"'


def test_reference_ancestor_package_from_nested_child():
    imports = set()
    name = get_type_reference(
        package="package.ancestor.nested.child",
        imports=imports,
        source_type="package.ancestor.Message",
    )

    assert imports == {"from .... import ancestor as ___ancestor__"}
    assert name == '"___ancestor__.Message"'


def test_reference_root_package_from_child():
    imports = set()
    name = get_type_reference(
        package="package.child", imports=imports, source_type="Message"
    )

    assert imports == {"from ... import Message as __Message__"}
    assert name == '"__Message__"'


def test_reference_root_package_from_deeply_nested_child():
    imports = set()
    name = get_type_reference(
        package="package.deeply.nested.child", imports=imports, source_type="Message"
    )

    assert imports == {"from ..... import Message as ____Message__"}
    assert name == '"____Message__"'


def test_reference_unrelated_package():
    imports = set()
    name = get_type_reference(package="a", imports=imports, source_type="p.Message")

    assert imports == {"from .. import p as _p__"}
    assert name == '"_p__.Message"'


def test_reference_unrelated_nested_package():
    imports = set()
    name = get_type_reference(package="a.b", imports=imports, source_type="p.q.Message")

    assert imports == {"from ...p import q as __p_q__"}
    assert name == '"__p_q__.Message"'


def test_reference_unrelated_deeply_nested_package():
    imports = set()
    name = get_type_reference(
        package="a.b.c.d", imports=imports, source_type="p.q.r.s.Message"
    )

    assert imports == {"from .....p.q.r import s as ____p_q_r_s__"}
    assert name == '"____p_q_r_s__.Message"'


def test_reference_cousin_package():
    imports = set()
    name = get_type_reference(package="a.x", imports=imports, source_type="a.y.Message")

    assert imports == {"from .. import y as _y__"}
    assert name == '"_y__.Message"'


def test_reference_cousin_package_different_name():
    imports = set()
    name = get_type_reference(
        package="test.package1", imports=imports, source_type="cousin.package2.Message"
    )

    assert imports == {"from ...cousin import package2 as __cousin_package2__"}
    assert name == '"__cousin_package2__.Message"'


def test_reference_cousin_package_same_name():
    imports = set()
    name = get_type_reference(
        package="test.package", imports=imports, source_type="cousin.package.Message"
    )

    assert imports == {"from ...cousin import package as __cousin_package__"}
    assert name == '"__cousin_package__.Message"'


def test_reference_far_cousin_package():
    imports = set()
    name = get_type_reference(
        package="a.x.y", imports=imports, source_type="a.b.c.Message"
    )

    assert imports == {"from ...b import c as __b_c__"}
    assert name == '"__b_c__.Message"'


def test_reference_far_far_cousin_package():
    imports = set()
    name = get_type_reference(
        package="a.x.y.z", imports=imports, source_type="a.b.c.d.Message"
    )

    assert imports == {"from ....b.c import d as ___b_c_d__"}
    assert name == '"___b_c_d__.Message"'


@pytest.mark.parametrize(
    ["full_name", "expected_output"],
    [
        ("package.SomeMessage.NestedType", ("package", "SomeMessage.NestedType")),
        (".package.SomeMessage.NestedType", ("package", "SomeMessage.NestedType")),
        (".service.ExampleRequest", ("service", "ExampleRequest")),
        (".package.lower_case_message", ("package", "lower_case_message")),
    ],
)
def test_parse_field_type_name(full_name, expected_output):
    assert parse_source_type_name(full_name) == expected_output



import pytest

from tests.output_aristaproto.nestedtwice import (
    Test,
    TestTop,
    TestTopMiddle,
    TestTopMiddleBottom,
    TestTopMiddleEnumBottom,
    TestTopMiddleTopMiddleBottom,
)


@pytest.mark.parametrize(
    ("cls", "expected_comment"),
    [
        (Test, "Test doc."),
        (TestTopMiddleEnumBottom, "EnumBottom doc."),
        (TestTop, "Top doc."),
        (TestTopMiddle, "Middle doc."),
        (TestTopMiddleTopMiddleBottom, "TopMiddleBottom doc."),
        (TestTopMiddleBottom, "Bottom doc."),
    ],
)
def test_comment(cls, expected_comment):
    assert cls.__doc__ == expected_comment
